# Utilities for printing and logging status and debug output. #
#
# $Source: C:/bin/GJS/RCS/Logging.pm,v $
# $Revision: 3.13 $    $Date: 2021-02-03 19:57:31-05 $	  $Author: gjs $
#

package GJS::Logging;	# Originally derived from package HsUtilities.pm.
my ($VERSION)			 = (q$Revision: 3.13 $ =~ /\d[\d\.\+]+/g);
my ($MODDATE, $MODTIME) = (q$Date: 2021-02-03 19:57:31-05 $ =~ /\d[-:\d]+/g);

use strict;
use warnings;

my $debug = 0;	# Controls package debug output.
$debug = 1  if grep(/^-d/i, @ARGV);	# Preliminary setting until set_debug below.

$debug > 0 && bugout("Version", version_info());
$debug > 1 && bugout("debug",	$debug);

our @ISA		= qw( Exporter );
our @EXPORT		= qw( bugout VERSION );
our @EXPORT_OK	= qw(
					set_debug  set_quiet  set_logfile  
					version_info  show_globals  
					REPORT  NOTE  WARN  ERROR  MESSAGE  LOG  bugout 
					elapsed_time  commas
				);
our %EXPORT_TAGS = (DEFAULT => [ qw( &bugout ) ],
					ALL	 => [ @EXPORT_OK ],
				 );

use Data::Dumper;		# For debug output of arrays and hashes.
# Default Dumper settings:
	$Data::Dumper::Indent = 1;		# Simple, compact indent.
	## $Data::Dumper::Indent = 2;  	# Indent by key length; lines can get long.
	## $Data::Dumper::Indent = 3;  	# Like 2, but includes array indexes.
	$Data::Dumper::Pad = "   ";		# Prefix for every line.
	$Data::Dumper::Useqq = 0;		# "" around strings, \t for tabs, ....
	$Data::Dumper::Quotekeys = 0;	# No quotes around simple hash keys.
	$Data::Dumper::SortKeys = 1;	# Sort hash tables by key.
	## $Data::Dumper::SortKeys = sub { [sort {$a cmp $b} keys %{$_[0]}] };
	## $Data::Dumper::SortKeys = \&dump_order;

# Can't get hash sorting to work!:
sub dump_order {
	my ($hash) = @_;
	return [ (sort {$a cmp $b} keys %$hash) ];
	## my @keys = sort {$a cmp $b} keys %{$hash};
	## print STDERR "dump keys = " . @keys . "\n";
	## return \@keys;
}

# ------------------------------------------------------------------------------
# Find this script's filename and directory path:

my ($moduledir, $modulename);	# This module's directory path and filename.
BEGIN {		# Make these available at "compile-time", e.g. for 'use lib'.
	use File::Basename;
	$modulename = basename(__FILE__);
	$moduledir  =  dirname(__FILE__);
}

# -----------------------------------------------------------------------------
# Private Package Variables:

my $quiet = 0;			# 1 => Suppress console output.
my $logfile = "";		# File path for messages and debug output.

$debug > 1 && bugout("\n");
$debug > 1 && bugout("modulename", $modulename);
$debug > 1 && bugout("moduledir", $moduledir);
$debug > 1 && bugout("logfile", $logfile);
$debug > 1 && bugout("quiet", $quiet);

# -----------------------------------------------------------------------------
# Accessor functions for a caller to set private variables in this package:

sub set_debug	   { ($debug)	= @_ ;	$debug && bugout("debug",$_[0]) }
sub set_logfile	   { ($logfile)	= @_ ;	$debug && bugout("\nlogfile",$_[0]) }
sub set_quiet	   { ($quiet)	= @_ ;	$debug && bugout("quiet",$_[0]) }

sub set_indent_style {
	my ($indent_style) = @_;
	# 1 = Simple, compact indent by fixed amount.
	# 2 = Indent by length of keys.  More readable, but can produce long lines.
	# 3 = Like 2, but includes array indexes.
	if ($indent_style < 0 || $indent_style > 3) {
		die "$modulename - ERROR: Data::Dumper::Indent style must be 0, 1, 2, or 3.\n  Exit";
	}
	$Data::Dumper::Indent = $indent_style;
}

sub show_globals
{
	bugout("debug",		$debug);
	bugout("quiet",		$quiet);
	## bugout("verbose",		$verbose);
	bugout("logfile",	$logfile);
}
## show_globals();
## exit 1;


# -----------------------------------------------------------------------------
# Debug output for variable values, including nested arrays and hashes.  Prints
# to console and to log file (not to HS log).  Examples:
#
#	 bugout("Name", $scalar);	# Prints Name = '<value of scalar>'.
#	 bugout("Name", \@array);	# Prints Name = [<contents of array>].
#	 bugout("Name", \%array);	# Prints Name = {<contents of hash>}.
#	 bugout("\nName", "Value");	# Prints blank line, then Name = 'Value'.
#	 bugout("\n", "Message");	# Prints blank line, then Message (no '=').
#	 bugout("\nMessage");		# ''
#	 bugout("\n");				# Prints just a blank line.
#	 bugout("");				# ''
#	 bugout();					# ''

sub bugout
{
	my ($name, $var, $modulename) = @_;

	# Find caller's filename, subroutine name, and line number:
	my ($cfile, $cpkg, $csub, $cline) = caller_info();
	$cfile = $modulename  if $modulename;

	my $output = "";	# Line(s) to be printed.
	my $prefix = "";	# Printed before output.
	my $final_nl = "";	# Printed after array/hash.

	# If name is absent, empty, or just a newline, then print a blank line:
	if (@_ == 0 || ! $name || $name eq "\n") {
		$prefix = "\n";
		# If there is also a value (message), then print location & message:
		if ($var) {
			$output .= sprintf "%-30s  %s\n", "$cfile $cline $csub", $var;
		}
	} else {
		# If name starts with newline, print a blank line:
		if ($name =~ s/^\n//) {
			$prefix = "\n";
			$final_nl = "\n";
		}

		# Print source code location:
		$output .= sprintf "%-30s  ", "$cfile $cline $csub";

		if (@_ == 1) {
			# Just a message, no variable:
			$output .= "$name\n";
		} else {
			# Variable value:
			if (ref($var) =~ /ARRAY|HASH|CODE|REF|GLOB/) {
				# Print array or hash, including any nested structure:
				$output .= sprintf "%16s : %s\n", $name, ref($var);
				$output .= Data::Dumper->Dump([$var], ["$name"]);
				$output .= $final_nl;
			}
			else {
				# Single value:
				if ($var && $var =~ /^[\d.]+$/) {
					$output .= sprintf "%16s = %s\n", $name, $var;	 # Numeric
				} elsif (defined $var) {
					$output .= sprintf "%16s = '%s'\n", $name, $var; # String
				} else {
					$output .= sprintf "%16s = undef\n", $name;		 # Undefined
				}
			}
		}

	}

	print STDOUT "$prefix$output";
	LOG($prefix, $output)  if $logfile;
	return $output;		# For use by the caller.
}

sub bugout_test
{
	my $number = 123.456;
	my $string = "string value";
	## my @colors		= qw( red green blue 0 cyan magenta yellow );
	my @colors =
		( "red", "green", "blue", 111, "cyan", "magenta", "yellow", "", undef );
	my %young  = ( hair => "red",  eyes => "blue", mood => "good" );
	my %older  = ( hair => "gray", eyes => "gray", mood => "gray" );

	bugout();		# Blank line.
	bugout("Test of 'bugout()'");
	bugout("logfile", $logfile);
	bugout("\nBlank line and message");
	bugout("\n", "Blank line and message");
	bugout("undef", undef);
	bugout("number", $number);
	bugout("string", $string);
	bugout("\n");
	bugout("colors", join("' '", @colors));		# Small array on one line.
	bugout("colors", \@colors);
	bugout("");
	bugout("young", \%young);
	bugout();
	bugout("older", \%older);
}

# Uncomment to run test:
## bugout_test(); exit 2;

# -----------------------------------------------------------------------------
# Does nothing in quiet mode.  Otherwise, prints a message to STDOUT, and to the
# log file if specified.  $filename is optional; specify it in cases where the
# filename displays as something useless like "(eval)" or "(eval 2)".

## sub logprint
sub printall
{
	my ($message, $filename) = @_;

	our ($logfile, $quiet, $read_only);

	## my $debug = $main::debug  if ! $debug;
	## my $debug = 4;

	# Get caller info if not specified:
	($filename) = caller_info()  if ! $filename;

	$debug > 3 && bugout("message",		$message,		$modulename);
	$debug > 3 && bugout("filename",	$filename,		$modulename);
	$debug > 3 && bugout("logfile",		$logfile,		$modulename);
	$debug > 3 && bugout("quiet",		$quiet,			$modulename);
	$debug > 3 && bugout("read_only",	$read_only,		$modulename);

	return	if $quiet;	# NO OUTPUT.

	# Print to console:
	if (! $message || $message eq "\n") {
		print "\n";		# Blank line.
	} else {
		print "$filename  $message\n"
	}

	if ($logfile) {
		# Print to logfile:
		open (LOG, ">>", "$logfile")  or  print STDOUT
					"$filename - ERROR: Could not open '$logfile' ($!)\n";
		binmode(LOG, "encoding(UTF-8)"); # Avoid "Wide character in print" error.
		if (! $message || $message eq "\n") {  # Blank line.
			print LOG "\n";
		} else {
			my $date_time = date_time_string();	# Current local time, readable.
			$debug > 3 && bugout("date_time", $date_time);
			print LOG "$date_time  $filename  $message\n";
		}
		close LOG;
	}
}

# -----------------------------------------------------------------------------
# Prints a multi-line message to STDERR (unless $quiet mode) and to logfile:

sub REPORT
{
	my ($message, $reason) = @_;
	my $text = MESSAGE("", $message, $reason, 1);	# One more call level.
	print STDERR $text  if ! $quiet;

	LOG("", MESSAGE("", "$message", "", 1))			if $logfile;
	LOG("", MESSAGE("", "    ($reason)", "", 1))	if $logfile && $reason;
}

# -----------------------------------------------------------------------------
# Prints a multi-line note to STDERR (unless $quiet mode) and to logfile:

sub NOTE
{
	my ($message, $reason) = @_;
	my $text = MESSAGE("NOTE", $message, $reason, 1);	# One more call level.
	print STDERR $text  if ! $quiet;

	LOG("", MESSAGE("", "NOTE: $message", "", 1))	if $logfile;
	LOG("", MESSAGE("", "    ($reason)", "", 1))		if $logfile && $reason;
}

# -----------------------------------------------------------------------------
# Prints a multi-line WARNING message to STDERR and to log file:

sub WARN
{
	my ($message, $reason) = @_;
	my $text = MESSAGE("WARNING", $message, $reason, 1);  # One more call level.
	print STDERR $text;		# Even in quiet mode.

	LOG("", MESSAGE("", "WARNING: $message", "", 1))	if $logfile;
	LOG("", MESSAGE("", "    ($reason)", "", 1))		if $logfile && $reason;
}

# -----------------------------------------------------------------------------
# Logs an error message and returns a multi-line string suitable for use with
# die().  Examples:
#
#	die ERROR("Could not open '$filename'", $!);
#	die ERROR("I hate when that happens");

sub ERROR
{
	my ($message, $reason) = @_;

	LOG("", MESSAGE("ERROR", "$message", "", 1))	if $logfile;
	LOG("", MESSAGE("", "  ($reason)", "", 1))	if $logfile && $reason;

	if ($reason)	{ return "  ERROR: $message\n  ($reason)  Exit" }
	else			{ return "  ERROR: $message\n  Exit" }
}

# -----------------------------------------------------------------------------
# Returns a multi-line message (most callers should omit $level):

sub MESSAGE
{
	my ($severity, $message, $reason, $level) = @_;

	return "\n"  if ! $severity && ! $message && ! $reason;	# Blank line.

	$severity .= ": "  if $severity;	# For prefix to message.

	$level = 0  if ! $level;
	my ($filename, $package, @etc) = caller_info($level);
	## my $dot = "";
	my $dot = $message =~ /\.$/  ?  ''  :  '.'; 
	if ($reason)	{ return	"$filename - $severity$message\n" .
								"$filename     ($reason).\n" }
	else			{ return	"$filename - $severity$message$dot\n" }
}

# -----------------------------------------------------------------------------
# Adds a message to the log file:

sub LOG
{
	my ($prefix, $output) = @_;

	my $date_time = date_time_string();		# Current local time, readable.
	if (! open (LOG, ">>", "$logfile")) {
		my $badfile = $logfile;	# For use in error message.
		$logfile = '';	# Trying to log this error just causes more trouble.
		die ERROR("Could not open log file '$badfile'", $!);
	}
	binmode(LOG, "encoding(UTF-8)");	# Avoid "Wide character in print" error.
	if ($output) {
		print LOG "$prefix$date_time  $output";
	} else {
		print LOG "$prefix";
	}
	close LOG;
}

# -----------------------------------------------------------------------------
# Returns information about the caller (level 0) or the caller's caller (level
# 1), and so on.  In subroutine Foo(), caller_info() returns the caller of
# Foo().  caller_info(-1) returns information about the line of that call,
# which seems rarely useful.

sub caller_info
{
	my ($level) = @_;
	$level = 0  if ! $level;

	## my $debug = 3;

	my ($package, $filename, $line, $subroutine) = caller($level+1);
	(undef, undef, undef, $subroutine) = caller($level+2);		# CALLED FROM.

	$filename =~ s|.*[/\\]||;					# Omit directory/folder path.
	$subroutine = ""  if ! $subroutine;
	$subroutine =~ s/\w+:://;					# Omit "package::".
	$subroutine = ""  if $subroutine eq "(eval)"; # Top level of package.

	# This subroutine's caller info:
	## my ($cpkg, $cfile, $cline, $csub) = caller(0);
	## (undef, undef, undef, $csub) = caller(1);		# CALLED FROM.

	# DO NOT CALL BUGOUT() OR REPORT()? FROM HERE; THEY CAUSE MUTUAL RECURSION.
	$debug > 1 && printall("caller_info filename   = '$filename'",	 $modulename);
	$debug > 2 && printall("caller_info package    = '$package'",	 $modulename);
	$debug > 2 && printall("caller_info subroutine = '$subroutine'", $modulename);
	$debug > 2 && printall("caller_info line       =  $line",		 $modulename);

	# NOTE - Return order is different from caller().
	return ($filename, $package, $subroutine, $line);
}

# This test simulates an output routine such as bugout() or printall().  For
# level 0 (default), we want info about who called THIS subroutine.  Code on
# that line should say "caller_info_test()".
#
# Expected results:
#	Level -1	A line within this routine (caller_info_test).
#	Level  0	The line below which calls this routine.
#	Level  1	The line which "use"d this package (implicitly in BEGIN).

sub caller_info_test
{
	foreach my $level (-1 .. 2) {
		my ($filename) = Logging::caller_info($level);		# Just filename.
		my ($cfile, $cpkg, $csub, $cline) = caller_info($level); # Full info.
		printall("");
		printall("$level: filename = '$cfile'");
		printall("$level: cfile    = '$cfile'");
		printall("$level: cpkg     = '$cpkg'");
		printall("$level: csub     = '$csub'");
		printall("$level: cline    = '$cline'");
	}
}

# Uncomment to run test:
## caller_info_test(); exit 2;

# -----------------------------------------------------------------------------
# Returns a human-readable string containing date and time in 12 hour format.
# TBD - Alternatively, we could use POSIX::strftime("... %X ...") to show time
# in current locale format.

sub date_time_string
{
	my ($time) = @_;
	$time = time()	if ! $time;	# Current date/time if no arguments.

	# Split into components:
	my ($sec,$min,$hr,$mday,$mon,$yr,$wday,$yday,$isdst) = localtime($time);

	$mon++;				# 1 .. 12.
	$yr += 1900;		# 4-digit year.

	my $date_time = sprintf "%d/%d/%02d %2d:%02d:%02d %s",
		$mon, $mday, $yr,
		1 + ($hr+11) % 12, $min, $sec,	# 12-hour format.
		$hr > 11 ? "PM" : "AM";

	return $date_time;
}

#------------------------------------------------------------------------------
# Returns elapsed time in a readable string format.  Example:
#
#	my $start_time = time();
#	...
#	my $prep_time = time();
#	...
#	my $prep_duration = elapsed_time($start_time, $prep_time);
#	my $full_duration = elapsed_time($start_time);

sub elapsed_time
{
    my ($start, $finish) = @_;
    $finish = time()  if ! $finish;		# Default end time is now.

	my $elapsed = $finish - $start;		# Number of seconds.

	if    ($elapsed < 1)	 { return "<1 second"; }
	elsif ($elapsed == 1)	 { return "$elapsed second"; }
	elsif ($elapsed < 60)	 { return "$elapsed seconds"; }
	else {
		my ($sec, $min, $hour, @etc) = gmtime($elapsed);
		return sprintf ("%02d:%02d:%02d  (%d s)", $hour, $min, $sec, $elapsed);
	}
}

# -----------------------------------------------------------------------------
# Reformats a string by inserting comma delimiters in a number:

sub commas
{
	my ($number) = @_;

	# print "[$number  ";
	$number =~ s/(\d)(\d\d\d)\b/$1,$2/;	# Thousands.
	$number =~ s/(\d)(\d\d\d,)/$1,$2/;	# Millions.
	$number =~ s/(\d)(\d\d\d,)/$1,$2/;	# Billions.
	$number =~ s/(\d)(\d\d\d,)/$1,$2/;	# Trillions.
	$number =~ s/(\d)(\d\d\d,)/$1,$2/;	# Quadrillions.
	$number =~ s/(\d)(\d\d\d,)/$1,$2/;	# Quintillions.
	$number =~ s/(\d)(\d\d\d,)/$1,$2/;	# Oh come on!
	# print "$number\n";

	return $number;
}

# -----------------------------------------------------------------------------
# Returns this package's version info:

sub version_info
{
	my $modtime = $MODTIME;
	$modtime =~ s/:\d\d-?\d*$//;			# Remove seconds and GMT offset.
	return "$VERSION  ($MODDATE $modtime)";
}

# -----------------------------------------------------------------------------

1;		# Successful completion.

# vim:tw=80 sw=4 ts=4

